; USR() function to sort a BASIC string array.

; MIT License

; Copyright (c) 2024 Mark J. Reed
;
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
;
; The above copyright notice and this permission notice shall be included in all
; copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
; SOFTWARE.
;
.include "cx16.inc"
.include "cbm_kernal.inc"

usrpok          = $0311                ; put handler addr here
golden          = $0400                ; where we install it
GETADR          = $fe0c                ; parse float usr() arg into int in YA

; main routine copies the sort routine into golden RAM and sets the USR vector
; to point to it.
main:
                lda #<copy
                sta gREG::r0L
                lda #>copy
                sta gREG::r0H
                lda #<golden
                sta gREG::r1L
                lda #>golden
                sta gREG::r1H
                lda #<end_code
                sec
                sbc #<start_code
                sta gREG::r2L
                lda #>end_code
                sbc #>start_code
                sta gREG::r2H
                jsr MEMORY_COPY
                lda #<golden
                sta usrpok
                lda #>golden
                sta usrpok+1
                rts

.segment "ZEROPAGE"
ptr:            .res 2                 ; generic utility pointer
base:           .res 2                 ; address of basic array
count:          .res 2                 ; number of elements
g:              .res 1                 ; current index into gaps table
gap:            .res 2                 ; current gap (gaps[g])
i:              .res 2                 ; current index in outer loop
i_addr:         .res 2                 ; address of a[i]
j:              .res 2                 ; current index in inner loop
j_addr:         .res 2                 ; address of a[j]
k:              .res 2                 ; k = j - gap
k_addr:         .res 2                 ; address of a[k]

; 3-byte string descriptors: length + address of first char
temp1:          .res 3                 ; temporary string descriptor 1
temp2:          .res 3                 ; temporary string descriptor 2

.segment "CODE"
copy:                                  ;  copy starting here
.org golden                            ;  assemble for here
start_code:     jmp sort

; Marcin Ciura's gap sequence for Shell's sort
gaps:           .word 701, 301, 132, 57, 23, 10, 4, 1

sort:           ; get array base address (USR parameter)
                jsr GETADR
                sta base+1
                sty base

                ; subtract 2 to get address where size is stored
                tya
                sec
                sbc #$02
                sta ptr
                lda base+1
                sbc #$00
                sta ptr+1

                ; read the size (which is stored big-endian)
                ldy #$00
                lda (ptr),y
                sta count+1
                iny
                lda (ptr),y
                sta count


; outermost loop:  g=0 
                stz g

; gap = gaps[g]
outer_loop:     lda g
                asl
                tax
                lda gaps, x
                sta gap
                inx
                lda gaps,x
                sta gap+1

; for i in gap to count-1
                lda gap
                sta i
                lda gap+1
                sta i+1

middle_loop:    lda i+1
                cmp count+1
                bcc middle_body
                bne j_middle_done
                lda i
                cmp count
                bcc middle_body

j_middle_done:  jmp middle_done

middle_body:    ; compute the address of the i'th element
                lda base
                sta i_addr
                lda base+1
                sta i_addr+1
                ldx #$03

add_loop_i:     lda i_addr
                clc
                adc i
                sta i_addr
                lda i_addr+1
                adc i+1
                sta i_addr+1
                dex
                bne add_loop_i

                ; save that element in temp1
                ldy #0
copy_loop_i:    lda (i_addr),y
                sta temp1,y
                iny
                cpy #$03
                bcc copy_loop_i

                lda i
                sta j
                lda i+1
                sta j+1

; inner loop: while j >= gap
inner_loop:     lda j+1
                cmp gap+1
                bcc inner_done
                bne inner_body
                lda j
                cmp gap
                bcc inner_done

inner_body:     lda j                  ; k = j - gap
                sec
                sbc gap
                sta k
                lda j+1
                sbc gap+1
                sta k+1

                ; compute the address of the k'th element
                lda base
                sta k_addr
                lda base+1
                sta k_addr+1
                ldx #$03

add_loop_k:     lda k_addr
                clc
                adc k
                sta k_addr
                lda k_addr+1
                adc k+1
                sta k_addr+1
                dex
                bne add_loop_k

                ldy #$00
copy_loop_k:    lda (k_addr),y
                sta temp2,y
                iny
                cpy #$03
                bcc copy_loop_k

                jsr compare
                bcs inner_done

; to copy a[k] to a[j], we need the address of a[j]
                lda base
                sta j_addr
                lda base+1
                sta j_addr+1
                ldx #$03

add_loop_j:     lda j_addr
                clc
                adc j
                sta j_addr
                lda j_addr+1
                adc j+1
                sta j_addr+1
                dex
                bne add_loop_j

; do the copy a[j] = a[k]
                ldy #$00
copy_loop_j:    lda (k_addr),y
                sta (j_addr),y
                iny
                cpy #$03
                bcc copy_loop_j

; decrement j by gap and loop
                lda j
                sec
                sbc gap
                sta j
                lda j+1
                sbc gap+1
                sta j+1
                bra inner_loop

inner_done:

; finally, copy temp (our original a[i]) to a[j]. That means we need the
; address of a[j], which we might not have yet because the loop body may not
; have executed even once.  So we (again) compute the address of the j'th
; element:
                lda base
                sta j_addr
                lda base+1
                sta j_addr+1
                ldx #$03

add_loop_j2:    lda j_addr
                clc
                adc j
                sta j_addr
                lda j_addr+1
                adc j+1
                sta j_addr+1
                dex
                bne add_loop_j2

; and do the copy
                ldy #$00
copy_loop_j2:   lda temp1,y
                sta (j_addr),y
                iny
                cpy #$03
                bcc copy_loop_j2

                inc i
                bne j_middle_loop
                inc i+1
j_middle_loop:  jmp middle_loop

; if the gap isn't already down to 1, go back and repeat for the next gap size
middle_done:    lda gap
                cmp #$01
                beq all_done
                inc g
                jmp outer_loop

all_done:       rts

; compare the two strings whose descriptors are in temp1 and temp2.
; set flags as with CMP, with temp1 string treated as the accumulator:
;   Z set if strings are equal
;   Carry clear if string in temp1 sorts before string in temp2
;   Carry set if strings are equal or temp1 sorts after temp2
compare:        ldy #$00

compare_loop:   cpy temp1
                bcs compare_done
                cpy temp2
                bcs compare_done
                lda (temp1+1),y
                cmp (temp2+1),y
                bne not_equal
                iny
                bne compare_loop

compare_done:   ldy temp1
                cpy temp2
not_equal:      rts

end_code:
